"""
Implementation of a time-stepping scheme for a bouncing ball (1D) 
solving the conctact LCP formulated as constrained optimization problem:
Find dp that minimizes   cost = R*dp 
with momentum difference   dp = p(k+1) + rc*p(k) >= 0   note sign reversal on reflection
and dependent contact momentum    R = dp - h*f - (1+rc)*p(k) >= 0
""" 
import numpy as np
from scipy.optimize import minimize
from scipy.optimize import Bounds
import matplotlib.pyplot as plt

N = 1000    # number of time intervals
h = 0.01    # time step
T = (N+1)*h # total time
t = np.arange(0, T, h)  # time grid
rc = 0.9    # restitution coefficient in range   0...1   on physical grounds (except special cases)
m = 1       # mass
g = 10      # gravitational acceleration
q0 = 0.5    # initial position (greater zero)
v0 = 0      # initial velocity

y = np.zeros(N+1)  # position at n. time-point
p = np.zeros(N+1)  # momentum at n. time-point
F = np.zeros(N)    # integrated external non-contact force over n. time-step
R = np.zeros(N)    # integrated contact force over n. time-step

"""external force (here gravitation)"""
f = lambda tt: -m*g 

"""cost function with   x = p(k+1) + rc*p(k) """
cost = lambda x, F1,P0: x[0]*(x[0] - F1 - (1+rc)*P0)

"""derivative of cost function"""
cost_der = lambda x, F1,P0: np.array([2*x[0] - F1 - (1+rc)*P0])

"""bounds on momentum change (maximal impact velocity from conservation of energy)"""
dpbound = m*(1+rc)*np.sqrt(2*g*q0 + v0**2)   
dpbounds = Bounds(-1.1*dpbound, 1.1*dpbound)   # some margin (could be one-sided, since only reflections at bottom)

y[0]= q0 
p[0]= m*v0 
for n in range(N):
    y[n+1] = y[n] + h*p[n]/m
    F[n] = h*f(t[n+1]) 
    p[n+1] = p[n] + F[n]; 
    if y[n+1] <= 0:   # collision occurs        

        lcp_args = (F[n], p[n])   
        ineq_cons = {'type': 'ineq',
             'fun' : lambda x, F1, P0: np.array([x[0], x[0] - F1 - (1+rc)*P0]),
             'jac' : lambda x, F1, P0: np.array([[1], [1]]),
             'args': lcp_args}

        x0 = 0   # initial guess,   dp = 0   means p(k+1) = -rc*p(k)
        
        res = minimize(cost, 
                x0, 
                method='SLSQP', 
                jac=cost_der,
                args = lcp_args,
                constraints=ineq_cons, 
                options={'ftol': 1e-10, 'disp': False},
                bounds=dpbounds)
        
        dp = res.x
        p[n+1] = dp - rc*p[n] 
        R[n] = dp - F[n] - (1+rc)*p[n]
        print(x0, dp)

# plot with two y-axes
fig, ax1 = plt.subplots()
ycolor = 'tab:blue'
Rcolor = 'tab:red'

ax1.plot(t, y, color=ycolor)
ax1.set_xlabel('time')
ax1.set_ylabel('position', color=ycolor)
ax1.set_ylim([0, 0.7])
ax1.tick_params(axis='y', labelcolor=ycolor)

ax2 = ax1.twinx()  # instantiate a second axes that shares the same x-axis
ax2.plot(t[1:], R, color=Rcolor)
ax2.set_ylabel('contact momentum', color=Rcolor)  # we already handled the x-label with ax1
ax2.set_ylim([0, 7])
ax2.tick_params(axis='y', labelcolor=Rcolor)

fig.tight_layout()  # otherwise the right y-label is slightly clipped
plt.show()